package laboratoire4.strategies;

import laboratoire4.Action;
import laboratoire4.IPawn;
import laboratoire4.game.Game;
import laboratoire4.game.GameUtils;
import laboratoire4.pawns.PawnMovement;
import laboratoire4.pawns.PawnUtils;

import java.util.Collection;
import java.util.stream.Collectors;

public class WinningStrategy implements Strategy {
    private Action<IPawn> winningAction;

    @Override
    public EvaluationResult getNextMove(Game game) {
        if (winningAction == null) {
            return null;
        }

        IPawn winningPawn = winningAction.getPawn();
        return new EvaluationResult(winningPawn.getRow(), winningPawn.getCol(), winningAction.getMovement());
    }

    @Override
    public int getWeight(Game game) {
        winningAction = findImmediateWinningAction(game);
        if (winningAction == null) {
            // On cherche un chemin où on peut gagner quoi qu'il arrive
            winningAction = findWinningPath(game);
        }

        if (winningAction != null) {
            return WEIGHT_MAX;
        }

        return 0;
    }

    private static Action<IPawn> findImmediateWinningAction(Game game) {
        Collection<IPawn> nearPawns = GameUtils.getMaxPawnsAsStream(game.getBoard(), game.getPlayer())
                .filter(p -> p.getRow() + p.getDirection() == p.getPlayer().getGoal())
                .collect(Collectors.toList());

        for (IPawn pawn : nearPawns) {
            for (PawnMovement movement : PawnMovement.values()) {
                if (pawn.isMoveValid(game.getBoard(), movement)) {
                    return new Action<>(pawn, movement);
                }
            }
        }

        return null;
    }

    private Action<IPawn> findWinningPath(Game game) {
        for (IPawn pawn : GameUtils.getMaxPawns(game.getBoard(), game.getPlayer())) {
            if (PawnUtils.distanceFromGoal(pawn) > 4) {
                continue;
            }

            for (PawnMovement movement : PawnMovement.values()) {
                int nextRow = pawn.getRow() + pawn.getDirection();
                int nextCol = pawn.getCol() + movement.getMove();

                if (pawn.isMoveValid(game.getBoard(), movement) &&
                        !PawnUtils.canBeCaptured(game, nextRow, nextCol, game.getPlayer()) &&
                        hasWinningPath(game, pawn, nextRow, nextCol)) {
                    return new Action<>(pawn, movement);
                }
            }
        }

        return null;
    }

    private boolean hasWinningPath(Game game, IPawn initialPawn, int row, int col) {
        if (row == initialPawn.getPlayer().getGoal()) {
            return true;
        }

        for (PawnMovement movement : PawnMovement.values()) {
            int nextRow = row + initialPawn.getDirection();
            int nextCol = col + movement.getMove();

            if (!initialPawn.isMoveValid(game.getBoard(), movement, row, col)) {
                continue;
            }

            IPawn nextPawn = game.getBoard()[nextRow][nextCol];
            if (nextPawn == null && hasWinningPath(game, initialPawn, nextRow, nextCol)) {
                return true;
            }

            if (nextPawn != null && !PawnUtils.areSamePlayers(initialPawn, nextPawn) &&
                    !PawnUtils.canBeCaptured(game, row, col, initialPawn.getPlayer()) &&
                    hasWinningPath(game, initialPawn, nextRow, nextCol)) {
                return true;
            }
        }

        return false;
    }
}
